/*
    Copyright (C) HWPORT.COM
    All rights reserved.
    Author: JAEHYUK CHO <mailto:minzkn@minzkn.com>
*/

#if !defined(__KERNEL__)
# error need __KERNEL__ define !
#endif

#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>

#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/if_macvlan.h>

#include <net/netns/generic.h>
#include <net/addrconf.h>
#include <net/ip6_tunnel.h>

#include <uapi/linux/ethtool.h>

#if 0L
#include "br_private.h" /* bridge port 에 대한 정보 접근을 위해 필요 ("net/bridge/br_private.h") */
#endif

struct hwport_interface_iter {
	struct seq_net_private p;
	loff_t	pos;

	struct net *net;
	struct net_device *dev;
};

typedef struct __hwport_net_ts {
} hwport_net_t;

static unsigned int g_hwport_net_id __read_mostly;

static struct hwport_interface_iter *hwport_interface_get_idx(struct seq_file *seq, struct hwport_interface_iter *iter, loff_t pos)
{
 	if((iter->pos <= 0) || (pos < iter->pos) || (iter->dev == NULL)) {
		iter->pos = 0;
		/* 전체 net namespace 순회 방식 */
		iter->dev = NULL;
		/* first exist dev in net namespace */
		for(iter->net = list_entry_rcu((&net_namespace_list)->next, struct net, list);
				(&iter->net->list) != (&net_namespace_list);
				iter->net = list_entry_rcu(iter->net->list.next, struct net, list)) {
			iter->dev = first_net_device_rcu(iter->net);
			if(iter->dev) {
				break;
			}
		}
	}

	while(iter->dev != NULL) {
		if(pos <= iter->pos) {
			break;
		}
		++iter->pos;
		/* 전체 net namespace 순회 방식 */
		iter->dev = next_net_device_rcu(iter->dev);
		while(iter->dev == NULL) { /* next exist dev in net namespace */
			iter->net = list_entry_rcu(iter->net->list.next, struct net, list);
			if((&iter->net->list) == (&net_namespace_list)) {
				/* end of net */
				break;
			}
			iter->dev = first_net_device_rcu(iter->net);
		}
	}

	if ((iter->dev == ((struct net_device *)0)) || (pos != iter->pos)) { /* forget it */
		iter->pos = 0;
		return NULL;
	}

	return iter;
}

static void *hwport_interface_seq_start(struct seq_file *seq, loff_t *pos)
	__acquires(RCU)
{
	struct hwport_interface_iter *iter = seq->private;

	rtnl_lock();
	rcu_read_lock();
	iter->pos = *pos;
	iter->net = NULL;
	iter->dev = NULL;

	return hwport_interface_get_idx(seq, iter, *pos);
}

static void *hwport_interface_seq_next(struct seq_file *seq, void *v, loff_t *pos)
{
	struct hwport_interface_iter *iter = seq->private;

	return hwport_interface_get_idx(seq, iter, ++*pos);
}

static void hwport_interface_seq_stop(struct seq_file *seq, void *v)
	__releases(RCU)
{
	rcu_read_unlock();
	rtnl_unlock();
}

static int hwport_interface_seq_show(struct seq_file *seq, void *v)
{
	struct hwport_interface_iter *iter = v;
	struct net *net = iter->net;
	struct net_device *dev = iter->dev;
	struct net_device *master = netdev_master_upper_dev_get(dev);

	int s_is_netif_running;
	int s_is_netif_carrier_ok;
	int s_is_netif_low_up;

	int s_count;

	/* ASSERT_RTNL(); */

	dev_hold(dev);

	s_is_netif_running = netif_running(dev);
	s_is_netif_carrier_ok = netif_carrier_ok(dev);
	s_is_netif_low_up = s_is_netif_running && s_is_netif_carrier_ok;
	seq_printf(
		seq,
		"dev=%s,ifindex=%d,if_port=%u,ref=%d,type=0x%X,features=0x%lX,flags=0x%X,priv_flags=0x%X,low_up=%d,mtu=%u",
		netdev_name(dev),
		dev->ifindex,
		(unsigned int)dev->if_port,
		netdev_refcnt_read(dev), /* dev 의 reference count 는 percpu 변수이므로 이 함수를 통해서 합산값을 읽을 수 있습니다. */
		(unsigned int)dev->type,
		(unsigned long)dev->features, /* XXX_FEATURES */
		dev->flags /* IFF_XXXX */,
		dev->priv_flags /* IFF_XXXX */,
		s_is_netif_low_up,
		dev->mtu
	);
#if 0L
	do {
		const struct dev_ifalias *alias = rcu_dereference(dev->ifalias);
		if(alias) {
			seq_printf(
				seq,
				",alias=%s",
				alias->ifalias
			);
		}
	}while(0);
#elif 0L /* legacy kernel */
	if(dev->ifalias) {
		seq_printf(
			seq,
			",alias=%s",
			dev->ifalias
		);
	}
#endif

#if 0L
	if(/* s_is_netif_running (SFP module 이 없는 경우 지연이 발생) */ s_is_netif_low_up) {
		struct ethtool_cmd ecmd;

		if(__ethtool_get_settings(dev, &ecmd) >= 0) { /* ASSERT_RTNL() */
			__u32 s_link_speed;

			if(ecmd.autoneg != ((__u8)0xffu /* AUTONEG_UNKNOWN */)) {
				seq_printf(
					seq,
					",autoneg=%u",
					(unsigned int)ecmd.autoneg
				);
			}

			s_link_speed = ethtool_cmd_speed(&ecmd); /* (ecmd.speed_hi << 16) | ecmd.speed */
			if((s_link_speed > ((__u32)0u)) && (s_link_speed != ((__u32)0xffffu /* SPEED_UNKNOWN */))) {
				seq_printf(
					seq,
					",speed=%lu",
					(unsigned long)s_link_speed
				);
			}

			if(ecmd.duplex != ((__u8)0xffu /* DUPLEX_UNKNOWN */)) {
				seq_printf(
					seq,
					",duplex=%u",
					(unsigned int)ecmd.duplex
				);
			}
		}
	}
#endif

	if(master != ((struct net_device *)0)) { /* bond, bridge 등의 멤버인 경우 */
		seq_printf(
				seq,
				"%s%s",
				",master=",
				master->name
				);
	}

	do { /* lowerdev (real_dev) */
		const char cg_first_lowerdev[] = {
			",lowerdev="
		};

		s_count = 0;

#if defined(IFF_MACVLAN)
		if(netif_is_macvlan(dev)) { /* macvlan interface 임 (IFF_MACVLAN_PORT 는 macvlan 인터페이스의 lowerdev 에 설정됨) */
			struct macvlan_dev *vlan = netdev_priv(dev);

			if(vlan->lowerdev) {
				seq_printf(
						seq,
						"%s%s",
						(s_count == 0) ? (const char *)(&cg_first_lowerdev[0]) : "|",
						netdev_name(vlan->lowerdev)
				);
				++s_count;
			}
		}
#endif
#if defined(IFF_802_1Q_VLAN)
		if(dev->priv_flags & IFF_802_1Q_VLAN) { /* vlan interface 임 */
			struct net_device *real_dev = vlan_dev_real_dev(dev);

			if(real_dev) {
				seq_printf(
						seq,
						"%s%s",
						(s_count == 0) ? (const char *)(&cg_first_lowerdev[0]) : "|",
						netdev_name(real_dev)
				);
				++s_count;
			}
		}
#endif
#if defined(IFF_EBRIDGE)
		if(netif_is_bridge_master(dev)) { /* ethernet bridge interface 임 (IFF_BRIDGE_PORT 는 bridge 인터페이스의 lowerdev 에만 설정됨) */
			struct net_device *slave_dev;

			for_each_netdev_rcu(net, slave_dev) {
				if(dev != netdev_master_upper_dev_get(slave_dev)) {
					continue;
				}
				
				seq_printf(
						seq,
						"%s%s",
						(s_count == 0) ? (const char *)(&cg_first_lowerdev[0]) : "|",
						netdev_name(slave_dev)
				);
				++s_count;
			}
		}
#endif
#if defined(IFF_BONDING)
		if(netif_is_bond_master(dev) || netif_is_bond_slave(dev)) { /* bond interface 임 (IFF_BONDING 은 master/slave 모두 설정된다는 것에 유의) */
			struct net_device *slave_dev;

			for_each_netdev_rcu(net, slave_dev) {
				if(dev != netdev_master_upper_dev_get(slave_dev)) {
					continue;
				}
				
				seq_printf(
						seq,
						"%s%s",
						(s_count == 0) ? (const char *)(&cg_first_lowerdev[0]) : "|",
						netdev_name(slave_dev)
				);
				++s_count;
			}
		}
#endif
#if defined(ARPHRD_IPGRE)
		if(dev->type == ARPHRD_IPGRE) { /* IPv4 GRE interface 임 */
			struct ip_tunnel *t = netdev_priv(dev);
			int local_ifindex = t->parms.link;
			struct net_device *lowerdev;

			if(local_ifindex > 0) { /* interface 가 지정된 구성 */
				lowerdev = dev_get_by_index_rcu(net, local_ifindex);
			}
			else { /* interface 가 지정되지 않고 local IP로 찾아야 하는 구성 */
				struct net_device *tdev;

				lowerdev = NULL;
				for_each_netdev_rcu(net, tdev) {
					struct in_device *in_dev;

					if(tdev == dev) {
						continue;
					}

					in_dev = in_dev_get(tdev);
					if(in_dev != NULL) {
						struct in_ifaddr __rcu **ifap = NULL;
						struct in_ifaddr *ifa = NULL;

						for (ifap = &in_dev->ifa_list; (ifa = rcu_dereference(*ifap)) != NULL; ifap = &ifa->ifa_next) {
							if(t->parms.iph.saddr == ifa->ifa_local) {
								lowerdev = tdev;
								break;
							}
						}
		
						in_dev_put(in_dev);

						if(lowerdev) {
							break;
						}
					}   
				}
			}

			if(lowerdev) {
				seq_printf(
						seq,
						"%s%s",
						(s_count == 0) ? (const char *)(&cg_first_lowerdev[0]) : "|",
						netdev_name(lowerdev)
				);
				++s_count;
			}
		}
#endif
#if defined(ARPHRD_IP6GRE)
		if(dev->type == ARPHRD_IP6GRE) { /* IPv6 GRE interface 임 */
			struct ip6_tnl *t = netdev_priv(dev);
			int local_ifindex = t->parms.link;
			struct net_device *lowerdev;
			
			if(local_ifindex > 0) { /* interface 가 지정된 구성 */
				lowerdev = dev_get_by_index_rcu(net, local_ifindex);
			}
			else { /* interface 가 지정되지 않고 local IP로 찾아야 하는 구성 */
				struct net_device *tdev;

				lowerdev = NULL;
				for_each_netdev_rcu(net, tdev) {
					struct inet6_dev *idev;
					
					if(tdev == dev) {
						continue;
					}
					
					idev = in6_dev_get(tdev);
					if(idev != NULL) {
						struct inet6_ifaddr *ifa6;

						read_lock_bh(&idev->lock);
						list_for_each_entry_rcu(ifa6, &idev->addr_list, if_list) {
							if(memcmp((const void *)(&t->parms.laddr), (const void *)(&ifa6->addr), sizeof(struct in6_addr)) == 0) {
								lowerdev = tdev;
								break;
							}
						}
						read_unlock_bh(&idev->lock);

						in6_dev_put(idev);
						
						if(lowerdev) {
							break;
						}
					}
				}
			}

			if(lowerdev) {
				seq_printf(
						seq,
						"%s%s",
						(s_count == 0) ? (const char *)(&cg_first_lowerdev[0]) : "|",
						netdev_name(lowerdev)
				);
				++s_count;
			}
		}
#endif
#if 0L && defined(IFF_ISATAP)
		if(dev->priv_flags & IFF_ISATAP) { /* isatap interface 임 */
		}
#endif
	}while(0);

	if(dev->addr_len > 0 /* exist mac ? */) {
		unsigned char s_dev_addr_offset;
		unsigned int hw_addr_index;
		struct netdev_hw_addr *hw_addr;

		/* all mac-list : 인터페이스가 가지고 있는 모든 MAC 주소를 순회 */
		hw_addr_index = 0u;
		netdev_hw_addr_list_for_each(hw_addr, &dev->dev_addrs) { /* 다음과 동일한 구현의 매크로 : list_for_each_entry(hw_addr, &dev->dev_addrs.list, list) */
			/* ignore zero mac */
			for(s_dev_addr_offset = 0u;s_dev_addr_offset < dev->addr_len;s_dev_addr_offset++) {
				if(hw_addr->addr[s_dev_addr_offset]) {
					break;
				}
			}
			if(s_dev_addr_offset >= dev->addr_len) {
				continue;
			}

			seq_printf(
					seq,
					"%s%02X",
					(hw_addr_index == 0u) ? ",HWaddr=" : "|",
					(unsigned int)hw_addr->addr[0]
					);
			for(s_dev_addr_offset = 1u;s_dev_addr_offset < dev->addr_len;s_dev_addr_offset++) {
				seq_printf(
						seq,
						":%02X",
						(unsigned int)hw_addr->addr[s_dev_addr_offset]
						);
			}

			++hw_addr_index;
		}
		if(unlikely(hw_addr_index == 0u /* no have list */ && dev->dev_addr /* but exist */)) { /* 만약 MAC 주소가 리스트에 없는 경우 추가 확인 */
			/* ignore zero mac */
			for(s_dev_addr_offset = 0u;s_dev_addr_offset < dev->addr_len;s_dev_addr_offset++) {
				if(dev->dev_addr[s_dev_addr_offset]) {
					break;
				}
			}
			if(s_dev_addr_offset < dev->addr_len) {
				seq_printf(
						seq,
						"%s%02X",
						(hw_addr_index == 0u) ? ",HWaddr=" : "|",
						(unsigned int)dev->dev_addr[0]
						);
				for(s_dev_addr_offset = 1u;s_dev_addr_offset < dev->addr_len;s_dev_addr_offset++) {
					seq_printf(
							seq,
							":%02X",
							(unsigned int)dev->dev_addr[s_dev_addr_offset]
							);
				}
			}
		}
	
		netif_addr_lock_bh(dev);

		/* unicast mac-list */
		hw_addr_index = 0u;
		netdev_hw_addr_list_for_each(hw_addr, &dev->uc) {
			/* ignore zero mac */
			for(s_dev_addr_offset = 0u;s_dev_addr_offset < dev->addr_len;s_dev_addr_offset++) {
				if(hw_addr->addr[s_dev_addr_offset]) {
					break;
				}
			}
			if(s_dev_addr_offset >= dev->addr_len) {
				continue;
			}

			seq_printf(
					seq,
					"%s%02X",
					(hw_addr_index == 0u) ? ",uc=" : "|",
					(unsigned int)hw_addr->addr[0]
					);
			for(s_dev_addr_offset = 1u;s_dev_addr_offset < dev->addr_len;s_dev_addr_offset++) {
				seq_printf(
						seq,
						":%02X",
						(unsigned int)hw_addr->addr[s_dev_addr_offset]
						);
			}

			++hw_addr_index;
		}

		/* multicast mac-list */
		hw_addr_index = 0u;
		netdev_hw_addr_list_for_each(hw_addr, &dev->mc) {
			/* ignore zero mac */
			for(s_dev_addr_offset = 0u;s_dev_addr_offset < dev->addr_len;s_dev_addr_offset++) {
				if(hw_addr->addr[s_dev_addr_offset]) {
					break;
				}
			}
			if(s_dev_addr_offset >= dev->addr_len) {
				continue;
			}

			seq_printf(
					seq,
					"%s%02X",
					(hw_addr_index == 0u) ? ",mc=" : "|",
					(unsigned int)hw_addr->addr[0]
					);
			for(s_dev_addr_offset = 1u;s_dev_addr_offset < dev->addr_len;s_dev_addr_offset++) {
				seq_printf(
						seq,
						":%02X",
						(unsigned int)hw_addr->addr[s_dev_addr_offset]
						);
			}

			++hw_addr_index;
		}
		
		netif_addr_unlock_bh(dev);
	}

#if 0L && defined(IFF_EBRIDGE)
	if(netif_is_bridge_master(dev)) { /* ethernet bridge interface 의 fdb mac 주소 */
		struct net_bridge *br = netdev_priv(dev);
		int s_fdb_index = 0;
		unsigned char s_dev_addr_offset;
		int i;

		for (i = 0; i < BR_HASH_SIZE; i++) {
			struct hlist_node *h;
			struct net_bridge_fdb_entry *f;

			hlist_for_each_entry_rcu(f, h, &br->hash[i], hlist) {
				seq_printf(
						seq,
						"%s%02X",
						(s_fdb_index == 0u) ? ",fdb=" : "|",
						(unsigned int)f->addr.addr[0]
						);
				for(s_dev_addr_offset = 1u;s_dev_addr_offset < ETH_ALEN;s_dev_addr_offset++) {
					seq_printf(
							seq,
							":%02X",
							(unsigned int)f->addr.addr[s_dev_addr_offset]
							);
				}
				if((f->is_local > 0) || (f->is_static > 0)) {
					seq_printf(
							seq,
							"?%s%s",
							f->is_local > 0 ? "L" : "",
							f->is_static > 0 ? "S" : ""
							);
				}
				++s_fdb_index;
			}
		}
	}
#endif

	do {/* IPv4 */
		struct in_device *in_dev;
		in_dev = in_dev_get(dev);
		if(in_dev != NULL) {
			struct in_ifaddr __rcu **ifap = NULL;
			struct in_ifaddr *ifa = NULL;

			for (s_count = 0, ifap = &in_dev->ifa_list; (ifa = rcu_dereference(*ifap)) != NULL; ifap = &ifa->ifa_next, s_count++) {
				seq_printf(
						seq,
						"%s%pI4/%u",
						s_count == 0 ? ",ipv4=" : "|",
						&ifa->ifa_local /* SIOCGIFADDR */,
						(unsigned int)ifa->ifa_prefixlen
						);
			}
		
			if(dev->flags & IFF_POINTOPOINT) { /* P-to-P's dst(remote) address */
				for (s_count = 0, ifap = &in_dev->ifa_list; (ifa = rcu_dereference(*ifap)) != NULL; ifap = &ifa->ifa_next, s_count++) {
					seq_printf(
							seq,
							"%s%pI4/%u",
							s_count == 0 ? ",dst_ipv4=" : "|",
							&ifa->ifa_address /* SIOCGIFDSTADDR */,
							(unsigned int)ifa->ifa_prefixlen
							);
				}
			}

			in_dev_put(in_dev);
		}   
	}while(0);

	do {/* IPv6 */
		struct inet6_dev *idev;
		idev = in6_dev_get(dev);
		if(idev != NULL) {
			struct inet6_ifaddr *ifa6;

			s_count = 0;
			read_lock_bh(&idev->lock);
			list_for_each_entry_rcu(ifa6, &idev->addr_list, if_list) {
				seq_printf(
						seq,
						"%s%pI6c/%u",
						s_count == 0 ? ",ipv6=" : "|",
						&ifa6->addr,
						(unsigned int)ifa6->prefix_len
						);

				++s_count;
			}
			read_unlock_bh(&idev->lock);

			in6_dev_put(idev);
		}
	}while(0);

	do {
		struct rtnl_link_stats64 s_stats64_local;

		if(dev_get_stats(dev, &s_stats64_local) != ((struct rtnl_link_stats64 *)0)) {
			seq_printf(
					seq,
					",rx_pkt=%llu,tx_pkt=%llu,rx_b=%llu,tx_b=%llu,rx_e=%llu,tx_e=%llu,rx_d=%llu,tx_d=%llu",
					(unsigned long long)s_stats64_local.rx_packets,
					(unsigned long long)s_stats64_local.tx_packets,
					(unsigned long long)s_stats64_local.rx_bytes,
					(unsigned long long)s_stats64_local.tx_bytes,
					(unsigned long long)s_stats64_local.rx_errors,
					(unsigned long long)s_stats64_local.tx_errors,
					(unsigned long long)s_stats64_local.rx_dropped,
					(unsigned long long)s_stats64_local.tx_dropped
					);
		}
	}while(0);

	seq_printf(seq, "\n");

	dev_put(dev);

	return 0;
}

static const struct seq_operations hwport_interface_seq_ops = {
	.start  = hwport_interface_seq_start,
	.next   = hwport_interface_seq_next,
	.stop   = hwport_interface_seq_stop,
	.show   = hwport_interface_seq_show,
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(5,0,0)
static int hwport_interface_seq_open(struct inode *inode, struct file *file)
{
	int rc;
	void *s_private;
	struct seq_file *seq;

	(void)inode;

	s_private = vzalloc(sizeof(struct hwport_interface_iter));
	if(unlikely(s_private == NULL)) {
		return -ENOMEM;
	}

	rc = seq_open(file, &hwport_interface_seq_ops);
	if(unlikely(rc < 0)) {
		vfree(s_private);
		return -ENOMEM;
	}

	seq = file->private_data;
	seq->private = s_private;

	return rc;
}

static int hwport_interface_seq_close(struct inode *inode, struct file *file)
{
	struct seq_file *seq = file->private_data;

	if(likely(seq->private)) {
		vfree(seq->private);
		seq->private = NULL;
	}

	return seq_release(inode, file);
}

static const struct file_operations hwport_interface_fops = {
	.owner = THIS_MODULE,
	.open = hwport_interface_seq_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = hwport_interface_seq_close,
};
#endif

static int __net_init hwport_net_pernet_init(struct net *net)
{
//	hwport_net_t *s_hwport_net = net_generic(net, g_hwport_net_id);
	int err;

	pr_info("initializing hwport_net pernet... (ns.inum=%u)\n", net->ns.inum);

#if LINUX_VERSION_CODE < KERNEL_VERSION(5,0,0)
	if(unlikely(proc_create("interface_info", S_IRUGO, net->proc_net, &hwport_interface_fops) == 0)) {
		printk("Create proc entry failed : interface_info\n");
		err = -ENOMEM;
		goto l_netdev_list_failed;
	}
#else /* Latest version */
	if(unlikely(proc_create_net("interface_info", S_IRUGO, net->proc_net, &hwport_interface_seq_ops, sizeof(struct hwport_interface_iter)) == 0)) {
		printk("Create proc entry failed : interface_info\n");
		err = -ENOMEM;
		goto l_netdev_list_failed;
	}
#endif

	pr_info("initialized hwport_net pernet... (ns.inum=%u)\n", net->ns.inum);

	return 0;
l_netdev_list_failed:;	
	return err;
}

static void __net_exit hwport_net_pernet_exit(struct net *net)
{
//	hwport_net_t *s_hwport_net = net_generic(net, g_hwport_net_id);

	pr_info("uninitializing hwport_net pernet... (ns.inum=%u)\n", net->ns.inum);

	remove_proc_entry("interface_info", net->proc_net);

	pr_info("uninitialized hwport_net pernet... (ns.inum=%u)\n", net->ns.inum);
}

static struct pernet_operations g_hwport_net_pernet = {
	/* struct list_head list; */
	.list = {}, /* LIST_HEAD_INIT(g_hwport_net_pernet.list), */
	/* int (*init)(struct net *net); */
	.init = hwport_net_pernet_init,
#if 0L
	/*void (*pre_exit)(struct net *net); */
	.pre_exit = NULL,
#endif
	/*void (*exit)(struct net *net); */
	.exit = hwport_net_pernet_exit,
#if 0L
	/*void (*exit_batch)(struct list_head *net_exit_list); */
	.exit_batch = NULL,
#endif

	/* unsigned int *id; */
	.id = (unsigned int *)(&g_hwport_net_id),
	/* size_t size; */
	.size = sizeof(hwport_net_t),
};

static int __init hwport_net_init(void)
{
	int err;

	pr_info("initializing hwport_net module...\n");

	/* int register_pernet_subsys(struct pernet_operations *); */
	err = register_pernet_subsys((struct pernet_operations *)(&g_hwport_net_pernet));
	if (unlikely(err < 0)) {
		pr_err("register pernet failed ! (g_hwport_net_pernet)\n");
		goto l_pernet_failed;
	}

	pr_info("initialized hwport_net module.\n");

	return 0;

l_pernet_failed:;	
	return err;
}

static void __exit hwport_net_exit(void)
{
	pr_info("uninitializing hwport_net module...\n");

	/* void unregister_pernet_subsys(struct pernet_operations *); */
	unregister_pernet_subsys((struct pernet_operations *)(&g_hwport_net_pernet));

	pr_info("uninitialized hwport_net module.\n");
}

module_init(hwport_net_init);
module_exit(hwport_net_exit);

MODULE_AUTHOR("JaeHyuk Cho <minzkn@minzkn.com>");
MODULE_DESCRIPTION("Linux kernel network hook");
MODULE_VERSION("hwport_net v0.0.1");
MODULE_LICENSE("GPL v2");

/* vim:set noexpandtab tabstop=4 shiftwidth=4 softtabstop=4 autoindent cindent smarttab fileencoding=utf8 fdm=marker: */
/* End of source */
